//	This code is adapted from the prototype of the paper "Real-Time Fluid Dynamics for Games" 
//	(J. Stam, Proc. GDC 2003)

#include <stdlib.h>
#include <math.h>
#include <GLUT/glut.h>
#include <iostream>
#include "UniformGrid.h"
#include "ColormapRenderer.h"
#include "solver.h"


using namespace std;

#define IX(i,j) ((i)+(N+2)*(j))


static int   N					= 200;		//Resolution of the simulation grid
static float dt					= 0.1;		//Timestep for the simulation
static float diff				= 0.000001;	//Diffusivity of the density. Large values mean less diffusion.
static float visc				= 0.00002;	//Viscosity of the fluid
static float force				= 1.0;		//Strength of force injected at mouse location
static float source			    = 40;		//Amount of denity injected at mouse lication

static float *u, *v, *dens;					//Simulation fields: velocity (u,v) and density (dens)
static float *u_prev, *v_prev, *dens_prev;	//Simulation values at previous timestep

static int win_x = 500, win_y = 500;
static int pressed = 0;						//State of the mouse button (pressed or not)
static int mx, my;							//Current location of mouse pointer
static int lmx,lmy;							//Previous location of mouse pointer (during mouse movement)
static bool frozen = false;					//Whether the simulation is running or nor


UniformGrid			grid(N+2,N+2,0,0,1,1);	//Used to visualize the density scalar field
ColormapRenderer*	renderer 	= 0;		//Used to render the grid, using color mapping
int					colormap    = 0;		//Colormap type for renderer


static void clear_data()					//Clear the entire simulation grid
{											//(initialize to zero values)
	int size=(N+2)*(N+2);
	for (int i=0 ; i<size ; ++i) 
		u[i] = v[i] = u_prev[i] = v_prev[i] = dens[i] = dens_prev[i] = 0.0f;
}

static void allocate_data()
{
	int size = (N+2)*(N+2);

	u			= new float[size];
	v			= new float[size];
	u_prev		= new float[size];
	v_prev		= new float[size];
	dens		= new float[size];
	dens_prev	= new float[size];
}



static void inject(float * d, float * u, float * v)		//Inject density and force on the line segment
{														//between the previous mouse position (lmx,lmy) and the
														//current mouse position (mx,my)
	float ndx=0, ndy=0;
	float dx = mx - lmx, dy = my - lmy;
	float len = sqrt(dx * dx + dy * dy);
	if (len != 0.0) {  ndx = dx / len; ndy = -dy / len; }
	
	float wn = float(win_x)/(N + 1);					// Grid cell width
	float hn = float(win_y)/(N + 1);					// Grid cell heigh

	int nsteps = len/((wn+hn)/2);
	if (nsteps==0) nsteps=1;
	float stepx = dx/nsteps;
	float stepy = dy/nsteps;

	for(int i=0;i<nsteps;++i)
	{
		float _mx = lmx + i*stepx;
		float _my = lmy + i*stepy;

		int i = (int)((       _mx /(float)win_x)*N+1);
		int j = (int)(((win_y-_my)/(float)win_y)*N+1);
		if (i<1 || i>N || j<1 || j>N) continue;

		float strength = ((float)(i+1))/nsteps;
		
		u[IX(i,j)] = force * strength * ndx;
		v[IX(i,j)] = force * strength * ndy;
		d[IX(i,j)] = source;
	}

	lmx = mx;											//Update the previous mouse position
	lmy = my;
}


static void key_func( unsigned char key, int x, int y )
{
	switch(key)
	{
		case 'c':										//Clear simulation grid
		case 'C':
			clear_data();
			break;
		case 'm':										//Change the colormap
		    colormap = (++colormap) % 3;					//Ask renderer to use another colormap
		    renderer->setColormapType((ColormapRenderer::COLORMAP_TYPE)colormap);
			break;	
		case ' ':
			frozen = !frozen;
			break;	
	}

	glutPostRedisplay();
}

static void mouse_click(int button, int state, int x, int y)
{
	pressed = (state == GLUT_DOWN);						//Remember mouse state
	if (pressed)										//If this is a click, reinitialize
	{													//the tracked mouse position
		lmx = mx = x;
		lmy = my = y;
	}	
}

static void mouse_drag( int x, int y )					//Track the current mouse position
{
	mx = x;
	my = y;
}

static void reshape_func( int width, int height )
{
	glutReshapeWindow(width, height);

	win_x = width;
	win_y = height;
}

static void do_one_simulation_step()					//Do one simulation time step
{
	if (frozen) return;									//If the simulation is stopped, nothing to do

	for(int i=0,size=(N+2)*(N+2); i<size ; ++i)			//Erase previous values
		dens_prev[i] = u_prev[i] = v_prev[i] = 0.0f;

	if (pressed)										//If mouse button is pressed, inject
	   inject(dens_prev, u_prev, v_prev);				//density and forces 

	vel_step(N, u, v, u_prev, v_prev, visc, dt);		//Solve for velocities
	dens_step(N, dens, dens_prev, u, v, diff, dt);		//Solve for densities

	glutPostRedisplay();								//Ask for a redraw
}

static void display_func()
{
	glViewport(0, 0, win_x, win_y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	ScalarAttributes& s = grid.pointScalars();			//Copy the simulation field 'dens' into the	
	for(int i=0,size=(N+2)*(N+2);i<size;++i)			//point scalars of our uniform grid. This
	   s.setC0Scalar(i,dens[i]);						//allows us to next display them.

	renderer->draw(grid);								//Render the grid (by color mapping its scalars)

	glutSwapBuffers();
}



int main(int argc, char ** argv)
{
	glutInit(&argc,argv);

	cout<<"6. 2D Fluid flow demo"<<endl<<endl;
	cout<<"In the graphics window, press"<<endl;
	cout<<"    c:        clear the flow field"<<endl;
	cout<<"    m:        cycle through different color maps"<<endl;
	cout<<"    space:    toggle animation on/off"<<endl;
	cout<<"Click the mouse to inject density"<<endl;
	cout<<"Click-and-drag the mouse to inject density and forces"<<endl;
	
	allocate_data();
	clear_data();

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowSize(win_x, win_y );
	glutCreateWindow( "6. 2D Fluid Flow");


    renderer = new ColormapRenderer();					//We must create the renderer AFTER OpenGL is fuly initialized.
														//This happens only after the GLUT window was created.
	renderer->setDrawingStyle(ColormapRenderer::DRAW_C1_TEX_CELLS);
														//Use texture-based color mapping (best quality)
	renderer->setColormapType((ColormapRenderer::COLORMAP_TYPE)colormap);

	glutKeyboardFunc(key_func);
	glutMouseFunc(mouse_click);
	glutMotionFunc(mouse_drag);
	glutReshapeFunc(reshape_func);
	glutIdleFunc(do_one_simulation_step);
	glutDisplayFunc(display_func);
	glutMainLoop();
}

